//  
//  Copyright (C) 2010 Rico Tzschichholz
// 
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
// 
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
// 
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
// 

using System;
using System.Collections.Generic;
using System.Linq;

using Cairo;
using Gdk;

namespace Docky.CairoHelper
{
	public class BufferCleanedEventArgs : EventArgs
	{
		public Gdk.Rectangle[] CleanedAreas { get; set; }
		
		public BufferCleanedEventArgs (Gdk.Rectangle[] areas)
		{
			CleanedAreas = areas;
		}
	}

	public class TileBufferedDockySurface : IDisposable
	{
		const int TILEWIDTH = 512, TILEHEIGHT = 512;
		const int MAXBUFFERDEPTH = 1;
		
		int BufferCount = 5;
		
		Dictionary<Gdk.Rectangle, DockySurface> tiles;
		List<Gdk.Rectangle> areas;
		
		public event EventHandler<BufferCleanedEventArgs> BufferCleaned;
		
		public Cairo.Surface Target { 
			get {
				if (tiles.Count > 0)
					return master.Context.Target;
				return null;
			}
		}
		public int Height { get; private set; }
		public int Width { get; private set; }
		
		DockySurface master;
		
		public TileBufferedDockySurface (int width_request, int height_request, Cairo.Surface target)
		{
			tiles = new Dictionary<Gdk.Rectangle, DockySurface> ();
			areas = new List<Gdk.Rectangle> ();
			master = new DockySurface (1, 1, target);
			
			Width = width_request;
			Height = height_request;

			int rows = height_request / TILEHEIGHT + 1;
			int cols = width_request / TILEWIDTH + 1;
			
			for (int row=0; row < rows; row++) {
				for (int col=0; col < cols; col++) {
					int width_left = width_request - (col * TILEWIDTH);
					int height_left = height_request - (row * TILEHEIGHT);
					areas.Add (new Gdk.Rectangle (col * TILEWIDTH, 
					                              row * TILEHEIGHT, 
					                              (width_request > TILEWIDTH ? (width_left > TILEWIDTH ? TILEWIDTH : width_left) : width_request), 
					                              (height_request > TILEHEIGHT ? (height_left > TILEHEIGHT ? TILEHEIGHT : height_left) : height_request)));
				}
			}
		}
		
		public void Draw (DockySurface source, int x, int y)
		{
			Gdk.Rectangle area = new Gdk.Rectangle (x, y, source.Width, source.Height);
			
			foreach (Gdk.Rectangle a in areas.Where (rect => rect.IntersectsWith (area))) {
				DockySurface destination;
				if (!tiles.TryGetValue (a, out destination)) {
					destination = new DockySurface (a.Width, a.Height, master);
					tiles.Add (a, destination);
				}
				source.Internal.Show (destination.Context, x - a.X, y - a.Y);
			}
			
			if (tiles.Count () > BufferCount) {
				area.Inflate (MAXBUFFERDEPTH * TILEWIDTH, MAXBUFFERDEPTH * TILEHEIGHT);
				KeepTiles (area);
			}
		}

		public DockySurface CreateSlice (Gdk.Rectangle area)
		{
			DockySurface result = new DockySurface (area.Width, area.Height, master);
			
			foreach (Gdk.Rectangle a in areas.Where (rect => rect.IntersectsWith (area))) {
				DockySurface source;
				if (tiles.TryGetValue (a, out source))
					source.Internal.Show (result.Context, a.X - area.X, a.Y - area.Y);
			}
			
			return result;
		}
		
		void KeepTiles (Gdk.Rectangle keepArea)
		{
			List<Gdk.Rectangle> removeAreas = new List<Gdk.Rectangle> ();
			removeAreas.AddRange (tiles.Keys.Where (a => !keepArea.IntersectsWith (a)));
			removeAreas.ForEach (a => {
				DockySurface s;
				tiles.TryGetValue (a, out s);
				if (s != null)
					s.Dispose ();
				tiles.Remove (a);
			});
			
			// Update count to avoid too much checks
			BufferCount = tiles.Count ();
			
			if (removeAreas.Any ())
				OnBufferCleaned (removeAreas.ToArray ());
		}
		
		void OnBufferCleaned (Gdk.Rectangle[] areas)
		{
			if (BufferCleaned != null)
				BufferCleaned (this, new BufferCleanedEventArgs (areas));
		}
		
		#region IDisposable implementation
		public void Dispose ()
		{
			BufferCleaned = null;
			foreach (DockySurface tile in tiles.Values) {
				(tile.Context.Target as IDisposable).Dispose ();
				tile.Dispose ();
			}
			tiles.Clear ();
			areas.Clear ();
			(master.Context.Target as IDisposable).Dispose ();
			master.Dispose ();
		}
		#endregion
	}
}

