//  
//  Copyright (C) 2010 Rico Tzschichholz
// 
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
// 
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
// 
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
// 

using System;
using System.Collections.Generic;
using System.Linq;

using Cairo;
using Gdk;
using Gtk;

using Docky.CairoHelper;
using Docky.Services;
using Docky.Services.Applications;

namespace Docky.Menus
{
	public class FileStackElement : StackElement
	{
		public GLib.File OwnedFile { get; private set; }
		GLib.File OwnedFileParent { get; set; }
		
		public bool IsReadOnly { get; private set; }
		public bool IsUnreadable { get; private set; }
		public bool IsSymbolic { get; private set; }
		
		public FileStackElement (GLib.File file)
		{
			OwnedFile = file;
			OwnedFileParent = OwnedFile.Parent;
			Label = OwnedFile.Basename;
		}
		
		#region implemented abstract members of StackElement
		protected override void UpdateInfo ()
		{
			//GLib.FileInfo info = OwnedFile.QueryInfo ("standard::icon,thumbnail::path,metadata::custom-icon," +
			//	"access::can-write,standard::is-symlink,access::can-read", GLib.FileQueryInfoFlags.None, null);
			//"metadata::emblems"
			
			//IsSymbolic = info.GetAttributeBoolean ("standard::is-symlink");
			//IsReadOnly = !info.GetAttributeBoolean ("access::can-write");
			//IsUnreadable = !info.GetAttributeBoolean ("access::can-read");
			
			IsSymbolic = OwnedFile.QueryInfo<bool> ("standard::is-symlink");
			IsReadOnly = !OwnedFile.QueryInfo<bool> ("access::can-write");
			IsUnreadable = !OwnedFile.QueryInfo<bool> ("access::can-read");

			FitIcon = false;
			
			string icon = null, label = null;

			if (OwnedFile.Basename.EndsWith (".desktop")) {
				DesktopItem d = new DesktopItem (OwnedFile);
				if (d.HasAttribute ("Icon"))
					icon = d.GetString ("Icon");
				if (d.HasAttribute ("Name"))
					Label = d.GetString ("Name");
				d.Dispose ();
			} else //if (!string.IsNullOrEmpty (icon = info.GetAttributeString ("metadata::custom-icon"))) {
			if (!string.IsNullOrEmpty (icon = OwnedFile.QueryInfo<string> ("metadata::custom-icon"))) {
				if (!icon.StartsWith ("file://"))
					icon = System.IO.Path.Combine (OwnedFile.StringUri (), icon);
			//} else if (!string.IsNullOrEmpty (icon = info.GetAttributeByteString ("thumbnail::path"))) {
			} else if (!string.IsNullOrEmpty (icon = OwnedFile.QueryInfo<string> ("thumbnail::path"))) {
				FitIcon = true;
			} else {
				//icon = DockServices.Drawing.IconFromGIcon (info.Icon);
				icon = OwnedFile.Icon ();
			}
			
			if (string.IsNullOrEmpty (icon))
				icon = "gtk-file";
			
			if (string.IsNullOrEmpty (label))
				label = OwnedFile.Basename;
			
			Icon = icon;
			Label = label;

			//info.Dispose ();
		}
		
		public override void DrawOnSurface (DockySurface surface)
		{
			if (!Accessed) {
				UpdateInfo ();
				Accessed = true;
			}
			
			// draw file-icon
			Gdk.Pixbuf icon_pbuf = IconPixbuf;
			bool is_icon_cached = (icon_pbuf != null);
			if (!is_icon_cached) {
				// buffer only IconTheme icons
				string icon_name = Icon;
				if (icon_name.Contains ('/')) {
					if (FitIcon)
						icon_pbuf = DockServices.Drawing.LoadIcon (icon_name, (int) (1.66 * StackWidget.IconSize), StackWidget.IconSize);
					else
						icon_pbuf = DockServices.Drawing.LoadIcon (icon_name, StackWidget.IconSize);
				} else if (!(is_icon_cached = StackWidget.icon_cache.TryGetValue (icon_name, out icon_pbuf)) || icon_pbuf == null) {
					icon_pbuf = DockServices.Drawing.LoadIcon (icon_name, StackWidget.IconSize);
					IconPixbuf = icon_pbuf;
					StackWidget.icon_cache.Add (icon_name, icon_pbuf);
					is_icon_cached = true;
				}
			}
			Gdk.CairoHelper.SetSourcePixbuf (surface.Context, icon_pbuf, 
			                                 (surface.Width - icon_pbuf.Width) / 2, 
			                                 (StackWidget.IconSize - icon_pbuf.Height) / 2 + StackWidget.Padding + StackWidget.LineWidth);
			surface.Context.Paint ();

			Gdk.Pixbuf emblem;
			string emblem_icon_name;
			if (IsSymbolic) {
				emblem_icon_name = "emblem-symbolic-link";
				if (!StackWidget.icon_cache.TryGetValue (emblem_icon_name, out emblem)) {
					emblem = DockServices.Drawing.LoadIcon (emblem_icon_name, StackWidget.IconSize / 3);
					StackWidget.icon_cache.Add (emblem_icon_name, emblem);
				}
				Gdk.CairoHelper.SetSourcePixbuf (surface.Context, emblem, 
				                                 (surface.Width + icon_pbuf.Width) / 2 - StackWidget.IconSize / 6, 
				                                 StackWidget.Padding + StackWidget.LineWidth);
				surface.Context.Paint ();
			}
			if (OwnedFileParent != null && OwnedFileParent.QueryInfo<bool> ("access::can-write") && IsReadOnly) {
				emblem_icon_name = "emblem-readonly";
				if (!StackWidget.icon_cache.TryGetValue (emblem_icon_name, out emblem)) {
					emblem = DockServices.Drawing.LoadIcon (emblem_icon_name, StackWidget.IconSize / 3);
					StackWidget.icon_cache.Add (emblem_icon_name, emblem);
				}
				Gdk.CairoHelper.SetSourcePixbuf (surface.Context, emblem, 
				                                 (surface.Width + icon_pbuf.Width) / 2 - StackWidget.IconSize / 6, 
				                                 2 * StackWidget.IconSize / 3 + StackWidget.Padding + StackWidget.LineWidth);
				surface.Context.Paint ();
			} else if (OwnedFileParent != null && OwnedFileParent.QueryInfo<bool> ("access::can-read") && IsUnreadable) {
				emblem_icon_name = "emblem-unreadable";
				if (!StackWidget.icon_cache.TryGetValue (emblem_icon_name, out emblem)) {
					emblem = DockServices.Drawing.LoadIcon (emblem_icon_name, StackWidget.IconSize / 3);
					StackWidget.icon_cache.Add (emblem_icon_name, emblem);
				}
				Gdk.CairoHelper.SetSourcePixbuf (surface.Context, emblem, 
				                                 (surface.Width + icon_pbuf.Width) / 2 - StackWidget.IconSize / 6, 
				                                 2 * StackWidget.IconSize / 3 + StackWidget.Padding + StackWidget.LineWidth);
				surface.Context.Paint ();
			}
			
			if (!is_icon_cached)
				icon_pbuf.Dispose ();

			// draw caption
			label_layout.SetText (Caption);
			label_layout.Width = Pango.Units.FromPixels (StackWidget.TextWidth);
			label_layout.FontDescription.AbsoluteSize = Pango.Units.FromPixels (StackWidget.TextHeight);
				
			surface.Context.MoveTo (StackWidget.LineWidth + StackWidget.Padding, 
			                        StackWidget.LineWidth + 2 * StackWidget.Padding + StackWidget.IconSize);
			Pango.CairoHelper.LayoutPath (surface.Context, label_layout);
			surface.Context.Color = DockMenu.IsLight ? new Cairo.Color (0.1, 0.1, 0.1, 0.9) : new Cairo.Color (1, 1, 1, 0.9);
			surface.Context.Fill ();
		}
		#endregion
		
		#region IDisposable implementation
		public override void Dispose ()
		{
			OwnedFile.Dispose ();
			OwnedFile = null;
			OwnedFileParent.Dispose ();
			OwnedFileParent = null;
			base.Dispose ();
		}
		#endregion
	}
}

