//  
//  Copyright (C) 2010 Rico Tzschichholz
// 
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
// 
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
// 
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
// 

using System;
using System.Collections.Generic;
using System.Linq;

using Cairo;
using Gdk;
using Gtk;

using Docky.CairoHelper;
using Docky.Services;

namespace Docky.Menus
{
	public abstract class StackElement : IDisposable
	{
		const uint ANIMATION_RATE = 32;
		
		public bool IsRealized { get; set; }
		
		protected bool Accessed { get; set; }
		
		public event EventHandler NeedsRedraw;
		
		public double HoverAnimationProgress { get; private set; }
		
		uint animation_timer;
		DateTime last_hovered = DateTime.UtcNow;
		bool hovered = false;
		public bool IsHovered {
			get { return hovered; }
			set {
				if (hovered == value)
					return;
				hovered = value;
				
				if ((DateTime.UtcNow - last_hovered).TotalMilliseconds > StackWidget.AnimationTime)
					last_hovered = DateTime.UtcNow;
				
				if (animation_timer > 0) 
					return;
				animation_timer = GLib.Timeout.Add (ANIMATION_RATE, delegate {
					QueueDraw ();
					
					double progress;
					if (hovered) {
						progress = Math.Min (1, (DateTime.UtcNow - last_hovered).TotalMilliseconds / StackWidget.AnimationTime);
						HoverAnimationProgress = progress;
					} else {
						progress = Math.Min (1, (DateTime.UtcNow - last_hovered).TotalMilliseconds / StackWidget.AnimationTime * 3);
						HoverAnimationProgress = 1 - progress;
					}

					if (progress < 1)
						return true;
	
					animation_timer = 0;
					return false;
				});
				QueueDraw ();
			}
		}
		
		public bool NeedsTooltip { get; private set; }
		
		int caption_truncate_offset = 0;
		public string Caption {
			get {
				if (!string.IsNullOrEmpty (Label) && Label.Length > 3 && caption_truncate_offset > 2 && NeedsTooltip)
					return Label.Substring (0, caption_truncate_offset - 2).TrimEnd () + "...";
				else
					return Label;
			}
		}
		
		public string Tooltip {
			get {
				return Label;
			}
		}
		
		string label;
		protected string Label { 
			get {
				return label;
			}
			set {
				label = value;
				label_height = null;
			}
		}
		
		public Gdk.Pixbuf IconPixbuf { get; set; }
	
		protected string Icon { get; set; }
		
		protected bool FitIcon { get; set; }
		
		public Gdk.Rectangle Area { get; set; }
		
		public int Height {
			get {
				return 2 * (StackWidget.Padding + StackWidget.LineWidth) + 
					StackWidget.IconSize + (LabelHeight > 0 ? StackWidget.Padding + LabelHeight : 0);
			}
		}
		
		public int Width {
			get {
				return StackWidget.DefaultElementWidth;
			}
		}
		
		int MaxCaptionLineCount;
		
		int? label_height;
		int LabelHeight {
			get {
				if (label_height.HasValue)
					return label_height.Value;
				
				if (string.IsNullOrEmpty (Label)) {
					label_height = 0;
					return label_height.Value;
				}
				
				label_layout.SetText (Label);
				label_layout.Width = Pango.Units.FromPixels (StackWidget.TextWidth);
				label_layout.FontDescription.AbsoluteSize = Pango.Units.FromPixels (StackWidget.TextHeight);
					
				caption_truncate_offset = 0;
				if (label_layout.LineCount > MaxCaptionLineCount) {
					NeedsTooltip = true;
					caption_truncate_offset = label_layout.Lines [MaxCaptionLineCount].StartIndex;
					label_layout.SetText (Caption);
				}
					
				Pango.Rectangle logical, ink;
				label_layout.GetPixelExtents (out ink, out logical);
				label_height = logical.Height;

				return label_height.Value;
			}
		}

		protected static Pango.Layout label_layout;
		
		static StackElement ()
		{
			label_layout = DockServices.Drawing.ThemedPangoLayout ();
			label_layout.Wrap = Pango.WrapMode.WordChar;
			label_layout.Alignment = Pango.Alignment.Center;
			label_layout.FontDescription = DockMenu.DefaultStyle.FontDescription;
			label_layout.Ellipsize = Pango.EllipsizeMode.None;
			label_layout.FontDescription.Weight = Pango.Weight.Normal;
		}
		
		public StackElement () : this (3)
		{
		}
		
		public StackElement (int labellines)
		{
			IsRealized = false;
			Accessed = false;
			NeedsTooltip = false;
			FitIcon = false;
			IconPixbuf = null;
			MaxCaptionLineCount = labellines;
		}
		
		public void MoveTo (int x, int y)
		{
			Area = new Gdk.Rectangle (x, y, Area.Width, Area.Height);
			IsRealized = false;
		}
		
		public void Reset ()
		{
			label_height = null;
			IsRealized = false;
			IconPixbuf = null;
		}
		
		public void QueueDraw ()
		{
			if (NeedsRedraw != null)
				NeedsRedraw (this, EventArgs.Empty);
		}
		
		protected abstract void UpdateInfo ();
		
		public abstract void DrawOnSurface (DockySurface surface);
		
		#region IDisposable implementation
		public virtual void Dispose ()
		{
			if (animation_timer > 0) 
				GLib.Source.Remove (animation_timer);
			NeedsRedraw = null;
			IconPixbuf = null;
		}
		#endregion
	}
}
