//  
//  Copyright (C) 2010 Rico Tzschichholz
// 
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
// 
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
// 
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
// 

using System;
using System.Collections.Generic;
using System.Linq;

using Cairo;
using Gdk;
using Gtk;

using Docky.CairoHelper;
using Docky.Services;

namespace Docky.Menus
{
	public enum StackScrollDirection
	{
		Vertical = 0,
		Horizontal,
	}
	
	public abstract class StackWidget : EventBox
	{
		public const int AnimationTime = 300;
		
		const int DefaultIconSize = 48;
		
		static int[] ZoomSteps = new int[] {33, 50, 66, 100, 150, 200, 300};
		const int DefaultZoomStep = 3;
		
		double ZoomChangedNotificationProgress = 0;
		
		uint zoom_changed_timer;
		int zoom_animation_in_steps = 20;
		int zoom_animation_out_steps = 60;
		int zoom_animation_counter = 0;
		
		static int current_zoom = DefaultZoomStep;
		int CurrentZoomStep {
			get {
				return current_zoom;
			}
			set {
				if (value < 0 || value >= ZoomSteps.Length)
					return;
				current_zoom = value;
				
				//FIXME force an even iconsize
				IconSize = DefaultIconSize * ZoomSteps [current_zoom] / 100;
				
				if (zoom_changed_timer > 0) 
					GLib.Source.Remove (zoom_changed_timer);
				
				zoom_animation_counter = 0;
				
				zoom_changed_timer = GLib.Timeout.Add (1000 / 60, delegate {
					zoom_animation_counter++;
					if (zoom_animation_counter <= zoom_animation_in_steps)
						ZoomChangedNotificationProgress = Math.Log (zoom_animation_counter) / Math.Log (zoom_animation_in_steps);
					else
						ZoomChangedNotificationProgress = 1 - Math.Log (zoom_animation_counter - zoom_animation_in_steps) / Math.Log (zoom_animation_out_steps);
					
					QueueDraw ();

					if (zoom_animation_counter < zoom_animation_in_steps + zoom_animation_out_steps)
						return true;
	
					zoom_changed_timer = 0;
					return false;
				});

				OnIconSizeChanged ();
			}
		}
		
		static int icon_size = DefaultIconSize;
		public static int IconSize {
			get {
				return icon_size;
			}
			private set {
				icon_size = value;

				TextWidth = 2 * icon_size;
				DefaultElementWidth = 2 * (icon_size + Padding + LineWidth);
				DefaultElementHeight = 2 * (Padding + LineWidth) + icon_size + Padding + TextHeight + TextLineSpacing;
			}
		}

		public static int LineWidth = 2;
		public static int Padding = 4;
		
		public static int TextHeight = 12;
		public static int TextLineSpacing = 3;
		public static int TextWidth = 2 * IconSize;

		public static int RowSpacing = 2;
		public static int ColumnSpacing = 2;
		
		public static int DefaultElementWidth = 2 * (IconSize + Padding + LineWidth);
		public static int DefaultElementHeight = 2 * (Padding + LineWidth) + IconSize + Padding + TextHeight + TextLineSpacing;


		public static Dictionary<string, Gdk.Pixbuf> icon_cache = new Dictionary<string, Gdk.Pixbuf> (25);
		
		public event EventHandler IconSizeChanged;
		
		List<StackElement> element_boxes = new List<StackElement> ();
		List<StackElement> visible_elements = new List<StackElement> ();
		
		TileBufferedDockySurface main_buffer;
		DockySurface visual_buffer;
		
		Gdk.Rectangle VisibleArea { get; set; }
		
		public Gtk.Adjustment Adjustment { get; private set; }
		
		uint scroll_animation_timer;
		DateTime last_scrolled;
		double start_position, target_position;
		double TargetPosition { 
			get {
				if (scroll_animation_timer == 0) {
					target_position = Adjustment.Value;
				}
				return target_position;
			}
			set {
				if (target_position == value)
					return;
				target_position = value;

				if (scroll_animation_timer == 0) {
					start_position = Adjustment.Value;
					last_scrolled = DateTime.UtcNow;
				}
				
				scroll_animation_timer = GLib.Timeout.Add (32, delegate {
					double progress = Math.Min (1, (DateTime.UtcNow - last_scrolled).TotalMilliseconds / AnimationTime);
					Adjustment.Value = start_position + (target_position - start_position) * (Math.Log ((progress * 9) + 1) / Math.Log (10));
					
					if (progress < 1)
						return true;
					
					scroll_animation_timer = 0;
					return false;
				});
			} 
		}
		
		protected double Position {
			get {
				return Adjustment.Value;
			}
		}
		
		public StackScrollDirection ScrollDirection { get; protected set; } 
		
		public int VisibleColumns { get; protected set; }
		public int VisibleRows { get; protected set; }
		
		public int Columns {
			get {
				int cols = 1;
				if (ScrollDirection == StackScrollDirection.Horizontal) {
					if (VisibleRows > 0)
						cols = Math.Max (1, (int) Math.Ceiling ((double) SourceElementsCount () / VisibleRows));
				} else {
					if (VisibleColumns > 0)
						cols = VisibleColumns;
					else
						cols = Math.Max (1, SourceElementsCount ());
				}
				return cols;
			}
		}

		public int Rows {
			get {
				int rows = 1;
				if (ScrollDirection == StackScrollDirection.Vertical) {
					if (VisibleColumns > 0)
						rows = Math.Max (1, (int) Math.Ceiling ((double) SourceElementsCount () / VisibleColumns));
				} else {
					if (VisibleRows > 0)
						rows = VisibleRows;
					else
						rows = Math.Max (1, SourceElementsCount ());
				}
				return rows;
			}
		}
		
		public int VisibleWidth {
			get {
				if (VisibleColumns > 0) {
					if (ScrollDirection == StackScrollDirection.Vertical && ColumnSizes != null && ColumnSizes.Length > 0)
						return Math.Max (ColumnSizes.Take (VisibleColumns).Sum () + ColumnSpacing * (VisibleColumns - 1) + 2 * Padding,
						                 DefaultElementWidth * VisibleColumns + ColumnSpacing * (VisibleColumns - 1) + 2 * Padding);
					else
						return DefaultElementWidth * VisibleColumns + ColumnSpacing * (VisibleColumns - 1) + 2 * Padding;
				} else {
					return FullWidth;
				}
			}
		}
		
		public int VisibleHeight {
			get {
				if (VisibleRows > 0) {
					if (ScrollDirection == StackScrollDirection.Horizontal && RowSizes != null && RowSizes.Length > 0)
						return Math.Max (RowSizes.Take (VisibleRows).Sum () + RowSpacing * (VisibleRows - 1) + 2 * Padding,
						                 DefaultElementHeight * VisibleRows + RowSpacing * (VisibleRows - 1) + 2 * Padding);
					else
						return DefaultElementHeight * VisibleRows + RowSpacing * (VisibleRows - 1) + 2 * Padding;
				} else {
						return FullHeight;
				}
			}
		}
		
		public int FullHeight {
			get {
				if (RowSizes == null || RowSizes.Length == 0)
					return 0;
				return RowSizes.Sum () + (RowSizes.Length - 1) * RowSpacing + 2 * Padding;
			}
		}

		public int FullWidth {
			get {
				if (ColumnSizes == null || ColumnSizes.Length == 0)
					return 0;
				return ColumnSizes.Sum () + (ColumnSizes.Length - 1) * ColumnSpacing + 2 * Padding;
			}
		}

		int[] RowSizes { get; set; }
		int[] ColumnSizes { get; set; }
		
		static Pango.Layout zoom_notification_label_layout;
		
		static StackWidget ()
		{
			zoom_notification_label_layout = DockServices.Drawing.ThemedPangoLayout ();
			zoom_notification_label_layout.Alignment = Pango.Alignment.Center;
			zoom_notification_label_layout.FontDescription = DockMenu.DefaultStyle.FontDescription;
			zoom_notification_label_layout.Ellipsize = Pango.EllipsizeMode.None;
			zoom_notification_label_layout.FontDescription.Weight = Pango.Weight.Bold;
		}
		
		public StackWidget (int visible_cols, int visible_rows, StackScrollDirection scroll_direction)
		{
			ScrollDirection = scroll_direction;
			
			if (visible_cols <= 0 && visible_rows <= 0) {
				switch (ScrollDirection) {
				case StackScrollDirection.Vertical:
					visible_rows = -1;
					visible_cols = 5;
					break;
				case StackScrollDirection.Horizontal:
					visible_rows = 4;
					visible_cols = -1;
					break;
				}
			}
			
			VisibleColumns = visible_cols;
			VisibleRows = visible_rows;
			
			VisibleWindow = false;
			AboveChild = true;
			
			Adjustment = new Adjustment (0, 0, 1, 0.01, 0.1, 0.1);
			Adjustment.ValueChanged += HandleAdjustmentValueChanged;
			
			AddEvents ((int) (Gdk.EventType.MotionNotify | Gdk.EventType.ButtonPress | Gdk.EventType.Scroll));

			//SetSizeRequest ();
		}

		void SetElements (List<StackElement> new_elements, bool dispose)
		{
			ResetBuffers ();
			
			visible_elements.Clear ();
			if (dispose)
				element_boxes.ForEach (e => {
					e.NeedsRedraw -= HandleElementNeedsRedraw;
					e.Dispose ();
				});
			element_boxes.Clear ();
			element_boxes = new_elements;
			
			SetSizeRequest ();
		}
		
		int[] row_sizes;
		int[] col_sizes;
		int[] row_offsets;
		int[] col_offsets;
		int col, row, cols, rows;
		
		List<StackElement> new_elements;
		
		protected void AddElement (StackElement new_element)
		{
			if (new_elements == null) {
				new_elements = new List<StackElement> (SourceElementsCount ());
				col = 0;
				row = 0;
				cols = Columns;
				rows = Rows;
				row_sizes = new int[rows];
				row_offsets = new int[rows];
				col_sizes = new int[cols];
				col_offsets = new int[cols];
			}
			
			new_element.Area = new Gdk.Rectangle (0, 0, new_element.Width, new_element.Height);
	
			col_sizes [col] = Math.Max (col_sizes [col], new_element.Area.Width);
			row_sizes [row] = Math.Max (row_sizes [row], new_element.Area.Height);

			col++;
			if (col >= cols) {
				col = 0;
				row++;
			}
			
			new_elements.Add (new_element);
		}
		
		protected void LayoutElements ()
		{
			LayoutElements (false);
		}

		protected void LayoutElements (bool is_relayout)
		{
			if (new_elements == null || !new_elements.Any ()) {
				SetElements (new List<StackElement> (), true);
				RowSizes = new int[0];
				ColumnSizes = new int[0];
				return;
			}
			
			int cols = Columns, rows = Rows;

			for (int i = 0; i < cols; i++)
				col_offsets [i] = Padding;
			for (int i = 0; i < cols - 1; i++)
				col_offsets [i + 1] = col_offsets [i] + ColumnSpacing + col_sizes [i];

			for (int i = 0; i < rows; i++)
				row_offsets [i] = Padding;
			for (int i = 0; i < rows - 1; i++)
				row_offsets [i + 1] = row_offsets [i] + RowSpacing + row_sizes [i];
			
			int col = 0, row = 0;
			foreach (StackElement element in new_elements) {
				element.MoveTo (col_offsets [col], row_offsets [row]);
				if (!is_relayout)
					element.NeedsRedraw += HandleElementNeedsRedraw;
				col++;
				if (col >= cols) {
					col = 0;
					row++;
				}
			}
			
			RowSizes = row_sizes;
			ColumnSizes = col_sizes;
			
			SetElements (new_elements, !is_relayout);
			new_elements = null;
		}
		
		void SetSizeRequest ()
		{
			if (ScrollDirection == StackScrollDirection.Vertical) {
				if (VisibleColumns > 0 && VisibleRows > 0)
					SetSizeRequest (VisibleWidth, VisibleHeight);
				else if (VisibleColumns <= 0)
					SetSizeRequest (VisibleWidth, FullHeight);
				else if (VisibleRows <= 0)
					SetSizeRequest (FullWidth, VisibleHeight);
				
				Adjustment.SetBounds (0, 1, 
				                      (double) (DefaultElementHeight + RowSpacing) / FullHeight,
				                      (double) VisibleHeight / FullHeight,
				                      (double) VisibleHeight / FullHeight);
			} else {
				if (VisibleColumns > 0 && VisibleRows > 0)
					SetSizeRequest (VisibleWidth, VisibleHeight);
				else if (VisibleColumns <= 0)
					SetSizeRequest (FullWidth, VisibleHeight);
				else if (VisibleRows <= 0)
					SetSizeRequest (VisibleWidth, FullHeight);
				
				Adjustment.SetBounds (0, 1, 
				                      (double) (DefaultElementWidth + ColumnSpacing) / FullWidth,
				                      (double) VisibleWidth / FullWidth,
				                      (double) VisibleWidth / FullWidth);
			}				
			
			if (Adjustment.Value != 0)
				Adjustment.Value = 0;
			else
				Adjustment.ChangeValue ();
			
			VisibleArea = Gdk.Rectangle.Zero;
			
			Console.WriteLine ("VC={0} VR={1} VW={2} VH={3} FW={4} FH={5} C={6} R={7} EC={8} DEW={9} DEH={10} IS={11}", 
			                   VisibleColumns, VisibleRows, VisibleWidth, VisibleHeight, FullWidth, FullHeight, Columns, Rows, 
			                   SourceElementsCount (), DefaultElementWidth, DefaultElementHeight, IconSize);
		}
		
		void HandleBufferCleaned (object sender, BufferCleanedEventArgs e)
		{
			foreach (Gdk.Rectangle area in e.CleanedAreas) {
				element_boxes.ForEach (box => {
					if (area.IntersectsWith (box.Area))
						box.IsRealized = false;
				});
			}
		}
		
		protected void HandleElementNeedsRedraw (object sender, EventArgs e)
		{
			QueueDraw ();
		}
		
		void HandleAdjustmentValueChanged (object sender, EventArgs e)
		{
			QueueDraw ();
		}
		
		protected override bool OnExposeEvent (EventExpose evnt)
		{
			using (Cairo.Context cr = Gdk.CairoHelper.Create (evnt.Window)) 
			{
				if (main_buffer == null) {
					main_buffer = new TileBufferedDockySurface (FullWidth, FullHeight, cr.Target);
					main_buffer.BufferCleaned += HandleBufferCleaned;
				}
				
				if (visual_buffer == null)
					visual_buffer = new DockySurface (Allocation.Width, Allocation.Height, main_buffer.Target);
				
				Gdk.Rectangle new_visible_area;
				if (ScrollDirection == StackScrollDirection.Vertical)
					new_visible_area = new Gdk.Rectangle (evnt.Area.X - Allocation.X, 
					                                      evnt.Area.Y - Allocation.Y + (int)(main_buffer.Height * Position), 
					                                      evnt.Area.Width, 
					                                      evnt.Area.Height);
				else
					new_visible_area = new Gdk.Rectangle (evnt.Area.X - Allocation.X + (int)(main_buffer.Width * Position), 
					                                      evnt.Area.Y - Allocation.Y, 
					                                      evnt.Area.Width, 
					                                      evnt.Area.Height);
				
				if (new_visible_area != VisibleArea) {
					VisibleArea = new_visible_area;
					
					// update visible elements and deselect old elements
					List<StackElement> old_visible_elements = visible_elements;
					visible_elements = element_boxes.FindAll (fb => fb.Area.IntersectsWith (VisibleArea));
					old_visible_elements.Except (visible_elements).ToList ().ForEach (e => e.IsHovered = false);
					
					// draw elements which are not-realized and visible in the current area
					visible_elements.FindAll (fb => !fb.IsRealized).ForEach (box => {
						using (DockySurface buffer = new DockySurface (box.Area.Width, box.Area.Height, main_buffer.Target))
						{
							box.DrawOnSurface (buffer);
							box.IsRealized = true;
							main_buffer.Draw (buffer, box.Area.X, box.Area.Y);
						}
					});
					
					visual_buffer.Clear ();
					DockySurface tmp = main_buffer.CreateSlice (VisibleArea);
					tmp.Internal.Show (visual_buffer.Context, 0, 0);
					tmp.Dispose ();

					DrawScrollableBordersFade (visual_buffer);
				}
				
				visual_buffer.Internal.Show (cr, Allocation.X, Allocation.Y);
				
				DockySurface overlay_buffer = new DockySurface (Allocation.Width, Allocation.Height, cr.Target);
				
				// draw element selections
				overlay_buffer.Context.Save ();
				overlay_buffer.Context.Operator = Operator.Add;
				visible_elements.FindAll (e => e.HoverAnimationProgress > 0).ForEach (e => {
					if (ScrollDirection == StackScrollDirection.Vertical)
						DrawElementSelection (overlay_buffer, new Gdk.Rectangle (e.Area.X + LineWidth, 
						                                                         e.Area.Y - (int)(main_buffer.Height * Position) + LineWidth, 
						                                                         e.Area.Width - 2 * LineWidth, e.Area.Height - 2 * LineWidth), 
						                      e.HoverAnimationProgress);
					else
						DrawElementSelection (overlay_buffer, new Gdk.Rectangle (e.Area.X - (int)(main_buffer.Width * Position) + LineWidth, 
						                                                         e.Area.Y + LineWidth, 
						                                                         e.Area.Width - 2 * LineWidth, 
						                                                         e.Area.Height - 2 * LineWidth), 
						                      e.HoverAnimationProgress);
				});
				overlay_buffer.Context.Restore ();
				
				// draw zoom-changed-notification
				if (ZoomChangedNotificationProgress > 0) {
					overlay_buffer.Context.Save ();
					overlay_buffer.Context.Operator = Operator.Over;

					zoom_notification_label_layout.SetText (string.Format ("{0}%", ZoomSteps [CurrentZoomStep]));
					zoom_notification_label_layout.Width = Pango.Units.FromPixels (Allocation.Width);
					zoom_notification_label_layout.FontDescription.AbsoluteSize = Pango.Units.FromPixels (72);
					Pango.Rectangle ink, logical;
					zoom_notification_label_layout.GetPixelExtents (out ink, out logical);
					
					overlay_buffer.Context.Color = DockMenu.IsLight ? new Cairo.Color (0.1, 0.1, 0.1, 0.4 * ZoomChangedNotificationProgress) : new Cairo.Color (1, 1, 1, 0.4 * ZoomChangedNotificationProgress);
					overlay_buffer.Context.RoundedRectangle ((Allocation.Width - ink.Width - 4 * Padding) / 2, (Allocation.Height - ink.Height - 4 * Padding) / 2, ink.Width + 8 * Padding, ink.Height + 8 * Padding, 16);
					overlay_buffer.Context.FillPreserve ();
					overlay_buffer.Context.LineWidth = 2 * LineWidth;
					overlay_buffer.Context.Color = !DockMenu.IsLight ? new Cairo.Color (0.1, 0.1, 0.1, 0.7 * ZoomChangedNotificationProgress) : new Cairo.Color (1, 1, 1, 0.7 * ZoomChangedNotificationProgress);
					overlay_buffer.Context.Stroke ();
					overlay_buffer.Context.MoveTo (0, (Allocation.Height - ink.Height - 4 * Padding) / 2);
					Pango.CairoHelper.LayoutPath (overlay_buffer.Context, zoom_notification_label_layout);
					overlay_buffer.Context.Fill ();
					
					overlay_buffer.Context.Restore ();
				}
				
				DrawScrollableBordersFade (overlay_buffer);

				overlay_buffer.Internal.Show (cr, Allocation.X, Allocation.Y);
				overlay_buffer.Dispose ();

				(cr.Target as IDisposable).Dispose ();
			}
			
			return true;
		}

		protected virtual void DrawElementSelection (DockySurface surface, Gdk.Rectangle area, double progress)
		{
			surface.Context.RoundedRectangle (area.X, area.Y, area.Width, area.Height, 6.0);
			surface.Context.Color = DockMenu.IsLight ? new Cairo.Color (0, 0, 0, progress * 0.1) : new Cairo.Color (1, 1, 1, progress * 0.1);
			surface.Context.FillPreserve ();
			surface.Context.LineWidth = LineWidth;
			surface.Context.Color = DockMenu.IsLight ? new Cairo.Color (0, 0, 0, progress * 0.5) : new Cairo.Color (1, 1, 1, progress * 0.5);
			surface.Context.Stroke ();
		}
		
		void DrawScrollableBordersFade (DockySurface surface)
		{
			// draw fade on top/bottom or left/right border if it is not reached
			if (ScrollDirection == StackScrollDirection.Vertical) {
				if (Allocation.Height < FullHeight) {
					surface.Context.Save();
					surface.Context.Operator = Operator.Source;
					surface.Context.Color = new Cairo.Color (0, 0, 0, 0);
	
					if (Position > 0.001) {
						Gradient lg = new LinearGradient (0, 0, 0, DefaultElementHeight / 3);
						lg.AddColorStop (0, new Cairo.Color (0, 0, 0, 1));
						lg.AddColorStop (1, new Cairo.Color (0, 0, 0, 0));
						surface.Context.Mask (lg);
						surface.Context.PaintWithAlpha (0);
						lg.Destroy ();
					}
	
					if (Position < 0.999 - (double) Allocation.Height / FullHeight) {
						Gradient lg = new LinearGradient (0, Allocation.Height - DefaultElementHeight / 3, 0, Allocation.Height);
						lg.AddColorStop (0, new Cairo.Color (0, 0, 0, 0));
						lg.AddColorStop (1, new Cairo.Color (0, 0, 0, 1));
						surface.Context.Mask (lg);
						surface.Context.PaintWithAlpha (0);
						lg.Destroy ();
					}
					
					surface.Context.Restore ();
				}
			} else {
				if (Allocation.Width < FullWidth) {
					surface.Context.Save();
					surface.Context.Operator = Operator.Source;
					surface.Context.Color = new Cairo.Color (0, 0, 0, 0);
	
					if (Position > 0.001) {
						Gradient lg = new LinearGradient (0, 0, DefaultElementWidth / 3, 0);
						lg.AddColorStop (0, new Cairo.Color (0, 0, 0, 1));
						lg.AddColorStop (1, new Cairo.Color (0, 0, 0, 0));
						surface.Context.Mask (lg);
						surface.Context.PaintWithAlpha (0);
						lg.Destroy ();
					}
	
					if (Position < 0.999 - (double) Allocation.Width / FullWidth) {
						Gradient lg = new LinearGradient (Allocation.Width - DefaultElementWidth / 3, 0, Allocation.Width, 0);
						lg.AddColorStop (0, new Cairo.Color (0, 0, 0, 0));
						lg.AddColorStop (1, new Cairo.Color (0, 0, 0, 1));
						surface.Context.Mask (lg);
						surface.Context.PaintWithAlpha (0);
						lg.Destroy ();
					}
					
					surface.Context.Restore ();
				}
			}
		}
		
		protected override bool OnLeaveNotifyEvent (EventCrossing evnt)
		{
			if (evnt.X < 0 || evnt.Y < 0 || evnt.X > Allocation.Width || evnt.Y > Allocation.Height) {
				visible_elements.FindAll (e => e.IsHovered).ForEach (e => e.IsHovered = false);
				TooltipText = "";
			}
			return base.OnLeaveNotifyEvent (evnt);
		}
		
		protected override bool OnMotionNotifyEvent (EventMotion evnt)
		{
			if (!visible_elements.Any ())
				return true;
			
			Gdk.Point spot;
			if (ScrollDirection == StackScrollDirection.Vertical)
				spot = new Gdk.Point ((int) evnt.X, (int) (evnt.Y + main_buffer.Height * Position));
			else
				spot = new Gdk.Point ((int) (evnt.X + main_buffer.Width * Position), (int) evnt.Y);
				
			StackElement hovered_box = visible_elements.Find (e => e.Area.Contains (spot));

			visible_elements.FindAll (e => e.IsHovered && e != hovered_box).ForEach (e => e.IsHovered = false);

			if (hovered_box == null) {
				TooltipText = "";
				return true;
			}

			if (!hovered_box.IsHovered) {
				hovered_box.IsHovered = true;		
				TooltipText = hovered_box.NeedsTooltip ? hovered_box.Tooltip : "";
			}
			
			return base.OnMotionNotifyEvent (evnt);
		}
		
		protected override bool OnButtonReleaseEvent (EventButton evnt)
		{
			Gdk.Point spot;
			if (ScrollDirection == StackScrollDirection.Vertical)
				spot = new Gdk.Point ((int) evnt.X, (int) (evnt.Y + main_buffer.Height * Position));
			else
				spot = new Gdk.Point ((int) (evnt.X + main_buffer.Width * Position), (int) evnt.Y);
			
			StackElement hovered_box = visible_elements.Find (e => e.Area.Contains (spot));

			if (hovered_box == null)
				return true;
			
			return OnElementClicked (hovered_box);
		}
		
		protected override bool OnScrollEvent (EventScroll evnt)
		{
			switch (evnt.Direction) {
			case Gdk.ScrollDirection.Down:
			case Gdk.ScrollDirection.Right:
				if ((evnt.State & ModifierType.ControlMask) == ModifierType.ControlMask)
					CurrentZoomStep--;
				else
					TargetPosition = Math.Max (Adjustment.Lower, Math.Min (Adjustment.Upper - Adjustment.PageSize, TargetPosition + Adjustment.StepIncrement));
				break;
			case Gdk.ScrollDirection.Up:
			case Gdk.ScrollDirection.Left:
				if ((evnt.State & ModifierType.ControlMask) == ModifierType.ControlMask)
					CurrentZoomStep++;
				else
					TargetPosition = Math.Max (Adjustment.Lower, Math.Min (Adjustment.Upper - Adjustment.PageSize, TargetPosition - Adjustment.StepIncrement));
				break;
			}
			return true;
		}
		
		protected void OnIconSizeChanged ()
		{
			if (IconSizeChanged != null)
				IconSizeChanged (this, EventArgs.Empty);
			
			VisibleArea = Gdk.Rectangle.Zero;			
			visible_elements.Clear ();

			// Mark elements for redraw, recalculate their position and relayout them
			element_boxes.ForEach (e => { e.Reset (); AddElement (e); } );
			LayoutElements (true);
			
			ResetBuffers ();
			ResetCaches ();
			
			SetSizeRequest ();
			
			QueueDraw ();
		}
		
		protected abstract bool OnElementClicked (StackElement element);
		
		public abstract int SourceElementsCount ();		
		
		protected override void OnHidden ()
		{
			Dispose ();
			base.OnHidden ();
		}		
		
		protected void ResetCaches ()
		{
			foreach (Gdk.Pixbuf pbuf in icon_cache.Values)
				pbuf.Dispose ();
			icon_cache.Clear ();
		}
		
		protected void ResetBuffers ()
		{
			if (main_buffer != null) {
				main_buffer.BufferCleaned -= HandleBufferCleaned;
				main_buffer.Dispose ();
			}
			main_buffer = null;

			if (visual_buffer != null)
				visual_buffer.Dispose ();
			visual_buffer = null;
		}
		
		public override void Dispose ()
		{
			if (IconSizeChanged != null)
				IconSizeChanged = null;
			
			ResetBuffers ();
			ResetCaches ();
			
			Adjustment.ValueChanged -= HandleAdjustmentValueChanged;
			Adjustment.Dispose ();
			
			if (scroll_animation_timer > 0)
				GLib.Source.Remove (scroll_animation_timer);
			if (zoom_changed_timer > 0) 
				GLib.Source.Remove (zoom_changed_timer);
			
			visible_elements.Clear ();
			element_boxes.ForEach (e => {
				e.NeedsRedraw -= HandleElementNeedsRedraw;
				e.Dispose ();
			});
			element_boxes.Clear ();
			
			base.Dispose ();
		}
	}
}

